from neau_hpc_utils.graph import plot_confusion_matrix
import timm
import torch
from torch import nn


class MyViT(nn.Module):
    """
    修改分类模型的分类数
    """
    def __init__(self, target_size, device, device_ids, pretrained=True):
        super(MyViT, self).__init__()
        self.model = timm.create_model('vit_base_patch16_224',
                                       pretrained=pretrained)
        n_features = self.model.head.in_features
        # 改成自己任务的图像类别数
        self.model.head = nn.Linear(n_features, target_size)
        self.model = torch.nn.DataParallel(self.model, device_ids=device_ids)
        self.model = self.model.to(device)

    def forward(self, x):
        x = self.model(x)
        return x


if __name__ == '__main__':
    y_pred = [0, 0, 2, 2, 0, 2]
    y_true = [2, 0, 2, 2, 0, 1]
    # array([[2, 0, 0],
    #        [0, 0, 1],
    #        [1, 0, 2]])
    cm = plot_confusion_matrix(y_true, y_pred,
                               normalize='all',
                               # display_labels=['a', 'b', 'c'],
                               # values_format='.3g',
                               # colorbar=False
                               # save_name='test3',
                               # save_path='../output'
                               )
    print(cm)



__author__ = ['yuquanfeng', 'zhiyaozhang']

import matplotlib.pyplot as plt
import os
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay
from sklearn.utils.multiclass import unique_labels


def plot_confusion_matrix(y_true,
                          y_pred,
                          save_name='',
                          postfix='jpg',
                          save_path='./',
                          normalize=None,
                          display_labels=None,
                          include_values=True,
                          title=None,
                          cmap=plt.cm.Blues,
                          xticks_rotation='horizontal',
                          figsize=(9, 7),
                          dpi=300,
                          values_format=None,
                          colorbar=True
                          ):
    """
    :param y_true:
    :param y_pred:
    :param normalize: {'true'， 'pred'， 'all'}， default=None
    :param save_name: 保存的文件名，默认值=’‘，为空字符串时，不保存文件
    :param postfix: 文件后缀名，('jpg', 'png')二选一，不符合要求时，默认jpg
    :param save_path: 文件保存的路径
    :param display_labels: 用于显示的label，默认值=None，不传入时，会自动计算出0~n的数字显示
    :param include_values: bool，默认值=True, 在混淆矩阵中包含值
    :param title: 标题, 默认值=None
    :param cmap: matplotlib识别的Colormap, 默认值=Blues
    :param xticks_rotation: {'vertical'，'horizontal'}或float, 默认值='horizontal', xtick标签的旋转。
    :param figsize: 默认值=(9, 7)
    :param dpi: 300
    :param values_format: str，默认值=None, 混淆矩阵中值的格式规范。如果“None”，格式规范为“d”或“.2g”，以较短者为准。
    :param colorbar: bool，默认值=True, 是否向绘图中添加颜色栏。
    :return: 混淆矩阵的值
    """
    cm = confusion_matrix(y_true, y_pred, normalize=normalize)
    _, ax = plt.subplots(figsize=figsize, dpi=dpi)

    if normalize:
        if not title:
            title = "Normalized confusion matrix"
    else:
        if not title:
            title = 'Confusion matrix, without normalization'

    if display_labels is None:
        display_labels = unique_labels(y_true, y_pred)

    disp = ConfusionMatrixDisplay(confusion_matrix=cm,
                                  display_labels=display_labels)
    disp.plot(include_values=include_values,
              cmap=cmap, ax=ax, xticks_rotation=xticks_rotation,
              values_format=values_format, colorbar=colorbar)

    plt.title(title)
    if postfix not in ('jpg', 'png'):
        postfix = 'jpg'

    if save_name and isinstance(save_name, str):
        if not os.path.exists(save_path):
            os.mkdir(save_path)
        plt.savefig(os.path.join(save_path, f'{save_name}.{postfix}'))
    plt.show()
    return cm


if __name__ == '__main__':
    y_pred = [0, 0, 2, 2, 0, 2]
    y_true = [2, 0, 2, 2, 0, 1]
    # array([[2, 0, 0],
    #        [0, 0, 1],
    #        [1, 0, 2]])
    cm = plot_confusion_matrix(y_true, y_pred,
                               normalize='all',
                               # display_labels=['a', 'b', 'c'],
                               # values_format='.3g',
                               # colorbar=False
                               save_name='test',
                               save_path='../output'
                               )
    print(cm)



__author__ = 'yuquanfeng'

import timm
import torch


def load_model_from_timm(model_name, device, device_ids, pretrained, num_classes):
    model = timm.create_model(model_name, pretrained=pretrained, num_classes=num_classes)
    model = model.to(device)
    model = torch.nn.DataParallel(model, device_ids=device_ids)
    return model




import torch
import cv2
import numpy as np
import pandas as pd
from torch.utils.data.dataset import Dataset
from torch.utils.data.dataloader import DataLoader
from torchvision.transforms.functional import resize
from .helper import pair


class ImgDataset(Dataset):
    def __init__(self, annotations_file, size=None, transform=None, target_transform=None):
        """
        :param annotations_file: 含有 "文件路径" 以及对应 "标签" 的csv文件名
                                eg:文件内容如下
                                    image, labels
                                    /xx/xxx/1.jpg, 0
                                    /xx/xxx/2.jpg, 1
        :param size: 对图片做resize的操作, 因为resize操作很常用，所以传入size后，直接对图片resize操作
                    支持传入整数，如 224， 也可列表[224, 224], 元组(224, 224)
        :param transform: 对图片做的变换
        :param target_transform: 对标签做的变换
        """
        self.img_label = pd.read_csv(annotations_file)
        self.size = pair(size)
        self.transform = transform
        self.target_transform = target_transform

    def __len__(self):
        return len(self.img_label)

    def __getitem__(self, idx):
        img_path = self.img_label.iloc[idx, 0]  # 获取图片的绝对路径
        img = cv2.imread(img_path)
        img = np.transpose(img, (2, 0, 1))  # 将shape变成（c, w, h）
        img = torch.tensor(img, dtype=torch.float)

        # 获得图片的分类
        label = self.img_label.iloc[idx, 1]

        # 对图片做变换
        if self.transform:
            img = self.transform(img)
        # 只有size满足 (int, int)或者 [int, int]时，才做resize变换
        if len(self.size) == 2 and isinstance(self.size[0], int) and isinstance(self.size[1], int):
            img = resize(img, size=self.size)

        # 对标签做变换
        if self.target_transform:
            label = self.target_transform(label)

        return img, label


def get_data_loader(annotations_file, size=None,
                    transform=None, target_transform=None,
                    shuffle=False, batch_size=32, num_workers=7):
    """
    读取图片数据集的label.csv文件，将会返回一个数据集的可迭代对象。
    :param annotations_file: 含有 "文件路径" 以及对应 "标签" 的csv文件名,eg：XXX.jpg,1
    :param size: 对图片做resize的操作时的size
    :param transform: 对图片做的变换
    :param target_transform: 对标签做的变换
    :param shuffle: bool
    :param batch_size: 批大小
    :param num_workers: 处理数据的线程数
    :return: 一个数据集的可迭代对象。
    """
    img_dataset = ImgDataset(annotations_file, size, transform, target_transform)
    img_dataloader = DataLoader(img_dataset, batch_size=batch_size,
                                num_workers=num_workers, shuffle=shuffle)
    return img_dataloader




__author__ = 'yuquanfeng'

"""
使用该库方法，需要传入数据集的根目录，存放图片资源的根目录，以及存放图片对应类别的csv文件，层级如下所示
--dataset\
----imgs\
------xxx.jpg
------subdir\
--------xxx.jpg
----label.csv

===================================
label.csv文件内容如下
image, labels
800113bb65efe69e.jpg, healthy

针对不同的label.csv文件，你需要继承DataLabel类之后，重写其_get_files_label()方法，
该方法返回(img_ids, label_ids, classes)元组

具体示例，参考 KagglePlantPathologyDataLabel类
===================================
具体使用示例，参考 get_split_data方法
"""

import os
import numpy as np
import pandas as pd
from tqdm import tqdm


class DataLabel:
    def __init__(self, files_dir, output_dir, label_csv_file_path, ratio: tuple = (7, 2, 1)):
        """
        初始化
        :param files_dir: 存放图片的根目录
        :param output_dir: 输出文件的根目录
        :param label_csv_file_path: 保存文件id和类别对应的文件路径
        :param ratio: 设置拆分比例，可选，默认(7, 2, 1)
        """
        self.files_dir = files_dir
        self.output_dir = output_dir
        self.label_csv_file_path = label_csv_file_path
        self.ratio = ratio
        self.paths = []  # 用来保存所有文件的绝对路径

    def split_dataset(self):
        """
        首先需要获取到文件的id列表img_ids，以及其对应标签label_ids，类别名称classes列表，
        请继承本类之后，重写get_files_label方法，该方法返回(img_ids, label_ids, classes)元组
        """
        print('-------------拆分数据集--------------')
        img_ids, label_ids, classes = self._get_files_label()
        train_idx, valid_idx, test_idx = self._get_random_idx(len(img_ids), self.ratio)
        train_df = self._get_random_dataset((img_ids, label_ids, classes), train_idx)
        valid_df = self._get_random_dataset((img_ids, label_ids, classes), valid_idx)
        test_df = self._get_random_dataset((img_ids, label_ids, classes), test_idx)

        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)

        train_df.to_csv(os.path.join(self.output_dir, 'train_label.csv'), index=False)
        valid_df.to_csv(os.path.join(self.output_dir, 'valid_label.csv'), index=False)
        test_df.to_csv(os.path.join(self.output_dir, 'test_label.csv'), index=False)
        print(f'-------------训练，验证，测试集的拆分长度为{len(train_df)},{len(valid_df)},{len(test_df)}--------------')

    def get_files_path(self, postfix=None):
        """
            通过传入文件根目录self.files_dir，可以找到该目录下所有的文件，返回文件路径列表。
            可选的postfix：指定扫描的文件后缀名

            :param: postfix: 传入文件后缀
            """
        print('-------------获取文件绝对路径列表--------------')
        result = []
        for root_path, dirs, files in os.walk(self.files_dir):
            for file in tqdm(files):
                if postfix and os.path.splitext(file)[1] == postfix:
                    result.append(os.path.join(root_path, file))
                else:
                    result.append(os.path.join(root_path, file))
        print(f'-------------共获取{len(result)}个{postfix}文件的绝对路径--------------')
        self.paths = result
        return result

    def _get_files_label(self):
        raise NotImplemented

    @staticmethod
    def _get_random_idx(dataset_size, ratio=(7, 2, 1)):
        """
        获取随机生成的索引列表
        :param dataset_size: 总的数据集的尺寸大小
        :param ratio: 拆分的比例
        :return: 训练集，验证集，测试集的随机索引列表
        """
        assert ratio[0] + ratio[1] + ratio[2] == 10, "总的比例之和必须等于10"
        train_split = ratio[0] * 0.1
        valid_split = ratio[1] * 0.1
        random_seed = 777

        indices = list(range(dataset_size))
        np.random.seed(random_seed)
        np.random.shuffle(indices)

        split1 = int(np.floor(train_split * dataset_size))
        split2 = int(np.floor((train_split + valid_split) * dataset_size))

        return indices[:split1], indices[split1:split2], indices[split2:]

    @staticmethod
    def _get_random_dataset(datasets, random_idx):
        img_ids, label_ids, classes = datasets
        img_ids, label_ids, classes = np.array(img_ids), np.array(label_ids), np.array(classes)
        train_img_ids, train_label_ids, train_classes = img_ids[random_idx], label_ids[random_idx], classes[random_idx]

        train_data = {'id': train_img_ids, 'label': train_label_ids, 'class_y': train_classes}
        return pd.DataFrame(train_data)


class KagglePlantPathologyDataLabel(DataLabel):
    def __init__(self, files_dir, output_dir, label_csv_file_path, ratio: tuple = (7, 2, 1)):
        super(KagglePlantPathologyDataLabel, self).__init__(files_dir, output_dir, label_csv_file_path, ratio)

    def _get_files_label(self):
        assert len(self.paths) > 0, "请先执行get_files_path()方法获取文件绝对路径列表"

        print(f'-----------获取文件的label-----------')
        label_df = pd.read_csv(self.label_csv_file_path)
        count = 0
        img_ids = []
        label_ids = []
        classes = []

        for idx in tqdm(range(len(self.paths))):
            path = self.paths[idx]
            img_name = path.split('/')[-1]
            df = label_df.loc[label_df['image'] == img_name]

            if len(df['labels']) == 1:
                category = df['labels'].item()
            else:
                continue

            if category == 'healthy':
                label = 0
            elif category == 'scab':
                label = 1
            elif category == 'frog_eye_leaf_spot':
                label = 2
            elif category == 'rust':
                label = 3
            elif category == 'complex':
                label = 4
            elif category == 'powdery_mildew':
                label = 5
            else:
                continue

            count += 1
            img_ids.append(path)
            label_ids.append(label)
            classes.append(category)

        print(f'-----------一共匹配了 {count} 个文件-----------')
        return img_ids, label_ids, classes


def get_split_data(SubClass,
                   root,
                   files_dir_name,
                   output_dir_name,
                   label_csv_name,
                   postfix=None,
                   ratio: tuple = (7, 2, 1)):
    """
    :param SubClass: DataLabel的子类
    :param root: 数据集根目录
    :param files_dir_name: 资源文件所在的根目录名称
    :param output_dir_name: 输出文件根目录名称
    :param label_csv_name: 文件id对应的标签的csv文件名称
    :param postfix: 文件后缀名，可选
    :param ratio: 设置拆分比例，可选，默认(7, 2, 1)
    """
    assert issubclass(SubClass, DataLabel), "SubClass必须继承自DataLabel类"

    # 1 设置文件根目录
    files_dir = os.path.join(root, files_dir_name)
    output_dir = os.path.join(root, output_dir_name)
    label_csv_path = os.path.join(root, label_csv_name)

    # 2 获取对象实例
    data_label = SubClass(files_dir, output_dir, label_csv_path, ratio)

    # 3 获取文件绝对路径列表
    data_label.get_files_path(postfix=postfix)

    # 4 将文件路径和其label一一对应，并拆分为训练，验证，测试三个子集，分别保存在csv文件中
    data_label.split_dataset()


if __name__ == '__main__':
    # get_split_data(SubClass=KagglePlantPathologyDataLabel,
    #                root='/stu01/datasets/Kaggle_plant_pathology/',
    #                files_dir_name='train_images',
    #                output_dir_name='output',
    #                label_csv_name='train.csv',
    #                postfix='.jpg'
    #                )
    """
    参考输出1：
    -------------获取文件绝对路径列表--------------
100%|██████████████████████████████████| 18632/18632 [00:00<00:00, 59746.39it/s]
100%|███████████████████████████████████████████| 1/1 [00:00<00:00, 9799.78it/s]
-------------共获取18633个.jpg文件的绝对路径--------------
-------------拆分数据集--------------
-----------获取文件的label-----------
100%|████████████████████████████████████| 18633/18633 [00:41<00:00, 448.52it/s]
-----------一共匹配了 17277 个文件-----------
-------------训练，验证，测试集的拆分长度为12093,3456,1728--------------
    """

    get_split_data(SubClass=KagglePlantPathologyDataLabel,
                   root='/stu01/datasets/Kaggle_plant_pathology/',
                   files_dir_name='train_images',
                   output_dir_name='output',
                   label_csv_name='train.csv',
                   postfix='.jpg',
                   ratio=(8, 1, 1)
                   )
    """
    参考输出2：
    -------------获取文件绝对路径列表--------------
100%|█████████████████████████████████| 18632/18632 [00:00<00:00, 349528.46it/s]
100%|██████████████████████████████████████████| 1/1 [00:00<00:00, 26546.23it/s]
-------------共获取18633个.jpg文件的绝对路径--------------
-------------拆分数据集--------------
-----------获取文件的label-----------
100%|████████████████████████████████████| 18633/18633 [00:27<00:00, 675.77it/s]
-----------一共匹配了 17277 个文件-----------
-------------训练，验证，测试集的拆分长度为13821,1728,1728--------------
    """



__author__ = 'yuquanfeng'

import csv
import os
import tqdm
import torch
import datetime
import pandas as pd


def pair(size):
    return size if isinstance(size, (list, tuple)) else (size, size)


def get_img_label_from_csv(csv_file_path):
    """
    读取{csv_file_path}文件内容如下
        image, labels
        /xx/xxx/1.jpg, 0
        /xx/xxx/2.jpg, 1
    的csv文件，返回图片路径列表和标签列表
    :param csv_file_path: csv文件的路径
    """
    with open(csv_file_path) as f:
        labels = []
        imgs = []
        csv_f = csv.reader(f)
        i = 0
        for row in csv_f:
            if i == 0:
                i += 1
                continue
            imgs.append(row[0])
            labels.append(int(row[1]))
    return imgs, labels


def get_imgs_path(root, postfix=None):
    """获取root目录下的所有文件路径，若postfix不为None，则只获取指定后缀的文件,如 “.jpg”
    """
    imgs_path = []
    for root, dirs, files in os.walk(root):
        for file in files:
            if not postfix:
                imgs_path.append(os.path.join(root, file))
            elif os.path.splitext(file)[1] == postfix:  # os.path.splitext() 将文件名和扩展名分开
                imgs_path.append(os.path.join(root, file))
    return imgs_path


class Accumulator:
    """For accumulating sums over `n` variables."""

    def __init__(self, n):
        self.data = [0.0] * n

    def add(self, *args):
        self.data = [a + float(b) for a, b in zip(self.data, args)]

    def reset(self):
        self.data = [0.0] * len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]


class LogTrainValid:
    """
    用于记录训练过程中的train和valid的精确率和损失值,
    并保存到csv文件中
    """

    def __init__(self):
        self.train_log = []
        self.train_acc_log = []
        self.train_loss_log = []
        self.val_acc_log = []
        self.val_loss_log = []

    def add(self, log, acc, loss, val_acc, val_loss):
        self.train_log.append(log)
        self.train_acc_log.append(acc)
        self.train_loss_log.append(loss)
        self.val_acc_log.append(val_acc)
        self.val_loss_log.append(val_loss)

    def save(self, filename):
        df = pd.DataFrame({'log': self.train_log,
                           'train_acc': self.train_acc_log,
                           'train_loss': self.train_loss_log,
                           'val_acc': self.val_acc_log,
                           'val_loss': self.val_loss_log})
        df.to_csv(filename, index=False)


class Classification:
    @staticmethod
    def train_loop(dataloader, model, loss_fn, optimizer, device):
        size = len(dataloader.dataset)
        num_batches = len(dataloader)
        model.train()
        test_loss, correct = 0, 0
        for X, y in tqdm(dataloader):
            X, y = X.to(device), y.to(device)

            # Compute prediction error
            pred = model(X)
            loss = loss_fn(pred, y)

            # save loss and correct per batch
            test_loss += loss.item()
            correct += (pred.argmax(1) == y).type(torch.float).sum().item()

            # Backpropagation
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        test_loss /= num_batches
        correct /= size
        print(f"Train: \n Accuracy: {(100 * correct):>0.1f}%, Avg loss: {test_loss:>8f}")

        return test_loss, correct

    @staticmethod
    def valid_loop(dataloader, model, loss_fn, device):
        size = len(dataloader.dataset)
        num_batches = len(dataloader)
        model.eval()
        test_loss, correct = 0, 0
        with torch.no_grad():
            for X, y in tqdm(dataloader):
                X, y = X.to(device), y.to(device)
                pred = model(X)
                test_loss += loss_fn(pred, y).item()
                correct += (pred.argmax(1) == y).type(torch.float).sum().item()
        test_loss /= num_batches
        correct /= size
        print(f"Valid: \n Accuracy: {(100 * correct):>0.1f}%, Avg loss: {test_loss:>8f} \n")

        return test_loss, correct


def get_current_time():
    now = datetime.datetime.now()
    return f'{now.year}-{now.month}-{now.day}-{now.hour}:{now.minute}'


